/*
 *	The Lestrade Search Library
 *
 *	(c) 2004--2005 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#ifndef _LESTRADE_H
#define _LESTRADE_H

#include <ucw/config.h>
#include <ucw/clists.h>
#include <custom/custom.h>

/* Global initialization and deinitialization of the library (init.c) */

void lestrade_init(byte *config_file);
void lestrade_cleanup(void);

/* Card source supplied by the user.  We have two types of data sources: enumerators
 * and fetchers. Enumerators get all available cards and for each call they return
 * a fastbuf with card data in tagged UTF-8 and fill in the supplied struct card_attrs.
 * Fetchers are given a struct card_attrs and return a fastbuf with card data.
 */

struct fastbuf;

typedef struct fastbuf * (*card_enumerator_t)(struct card_attrs *attrs, void *priv);
typedef struct fastbuf * (*card_fetcher_t)(struct card_attrs *attrs, void *priv);

/* Operations with indices (index.c and indexer.c) */

/* Constructs an index from the given card source, embeds custom_data in it and returns number of documents indexed */
uint lestrade_make_index(byte *filename, card_enumerator_t source, void *enum_priv, struct custom_index_hdr *custom_hdr, uint custom_hdr_len);

/* Load index into memory and free it */
struct lestrade_index;
int lestrade_open_index(byte *filename, struct lestrade_index **pidx);
void lestrade_close_index(struct lestrade_index *idx);
struct custom_index_hdr *lestrade_get_custom_index_hdr(struct lestrade_index *idx);

/* Queries and operations on them (query.c) */

struct lestrade_query {			/* This structure describes a single query */
  struct lestrade_expr *expr;		/* Query expression */
  int error_code;			/* LESTRADE_ERR_xxx */
  byte *error_msg;
  uint trace_mask;			/* Which TRACE_xxx messages should be logged */

  card_fetcher_t fetcher;		/* Called whenever a full text of a card is needed */
  void *fetcher_priv;			/* Private data passed to the fetcher */
  void *custom_priv;			/* For use by custom_match_card() and similar functions */

  struct lestrade_result *results;	/* Will be filled by lestrade_get_answers() */
  uint num_results;
  uint max_results;			/* How many results do you wish to return (default: 100) */

  struct query *q;			/* Private data here, no trespassing! */
  struct mempool *pool;			/* Pool from which all query-related structures are allocated */
};

enum lestrade_expr_type {		/* Types of lestrade_expr */
  EX_RESERVED,				/* Undefined node type */
  EX_MATCH,				/* Match word or phrase or simple-search term */
  EX_AND,				/* Boolean operators */
  EX_OR,
  EX_NOT,
  EX_ANY,				/* Logical 1 */
  EX_NONE,				/* Logical 0 */
  /* The following types are used internally */
  EX_IGNORE,				/* Tried to match an ignored word */
  EX_WORD,				/* Match a given word together with its expansions */
  EX_PHRASE,				/* Match a given phrase */
};

struct lestrade_atom {			/* A single element of a simple-search expression */
  cnode n;
  byte *word;				/* The word or phrase to be matched */
  u32 type_set;				/* Set of allowed word types (default: all) */
  int weight;				/* Q points per this word (default: 10000) */
  s8 sense;				/* Matching sense: -1=NO, 0=MAYBE, 1=YES (default: YES) */
  byte accent_mode;			/* ACCENT_MODE_xxx (default: STRICT) */
  byte morph_mode;			/* MORPH_MODE_xxx (default: OFF) */
  struct lestrade_expr *xlat;		/* [internal] translation of the atom */
  clist words;				/* [internal] struct word's extracted from this atom */
};

struct lestrade_expr {			/* Query expression */
  enum lestrade_expr_type type;
  union {
    struct {				/* Boolean operators */
      struct lestrade_expr *l, *r;
    } op;
    struct {				/* EX_MATCH */
      clist atoms;
    } match;
    struct word *word;			/* EX_WORD */
    struct phrase *phrase;		/* EX_PHRASE */
  } u;
};

enum lestrade_accent_mode {
  ACCENT_MODE_STRIP = 0,		/* Don't consider accents */
  ACCENT_MODE_STRICT = 1,		/* Always match accents */
};

enum lestrade_morph_mode {
  MORPH_MODE_OFF = 0,			/* No morphing */
  MORPH_MODE_ON = 1,			/* Morphing enabled, but generated forms are matched with strict accents */
  MORPH_MODE_POSTACCENT = 2,		/* Morphing enabled, generated forms are matched in the current accent mode */
  MORPH_MODE_PREACCENT = 3,		/* As above, but honor the accent mode even when doing the morphological analysis */
};

enum lestrade_error_code {
  LESTRADE_ERR_OK = 0,
  LESTRADE_ERR_ONLY_NONIDX = -100,	/* Simple search expression contains only non-indexed words */
  LESTRADE_ERR_WORD_TOO_LONG = -101,
  LESTRADE_ERR_TOO_MANY_WORDS = -102,
  LESTRADE_ERR_TOO_MANY_PHRASES = -103,
  LESTRADE_ERR_TOO_MANY_BOOLS = -104,
  LESTRADE_ERR_TOO_MANY_OCCURENCES = -105,
  LESTRADE_ERR_ANY_DOCUMENT_MATCHES = -106,
  LESTRADE_ERR_PHRASE_TOO_LONG = -107,
  LESTRADE_ERR_WILD_TOO_BROAD = -108,	/* Non-wildcard prefix of wildcard pattern shorter than WildMinPrefix */
  LESTRADE_ERR_WILD_TOO_LONG = -109,	/* Wildcard pattern too long */
  LESTRADE_ERR_WILD_TOO_RICH = -110, 	/* Wildcard pattern matched by too many words */
  LESTRADE_ERR_INDEX_NOT_FOUND = -200,	/* Errors which can be returned by lestrade_open_index() */
  LESTRADE_ERR_INDEX_STAT_FAILED = -201,
  LESTRADE_ERR_INDEX_MMAP_FAILED = -202,
  LESTRADE_ERR_INDEX_MALFORMED = -203,
  LESTRADE_ERR_INDEX_NOT_FINISHED = -204,
  LESTRADE_ERR_INDEX_BAD_VERSION = -205,
};

struct lestrade_result {		/* A single occurence of this structure per matching card */
  int quality;				/* Number of quality points for this match */
  struct card_attrs attrs;		/* Attributes of the matching card */
  struct custom_result custom;		/* Custom data set by custom_match_card() */
};

struct lestrade_matches {		/* Returned by lestrade_eval_matches() */
  byte *text;				/* Full text of the matched card */
  uint text_length;
  struct lestrade_word_pos *words;	/* Positions of all words found in the card */
  uint num_words;
  struct lestrade_match *matches;	/* Array describing all matches found in the card (sorted by position) */
  uint num_matches;
  int quality;				/* Total quality points received for this document */
};

struct lestrade_word_pos {
  uint pos;				/* Position in the text measured in bytes */
  byte length;				/* Number of bytes */
  byte word_type;			/* WT_xxx (0xff for a dummy word at the end of the text) */
  byte word_class;			/* WC_xxx (non-zero if not a normal word; 0xff if position doesn't correspond to a word) */
  byte word_index;			/* Index of the word inside the query (0xff if not matched) */
					/* If multiple query words match, an index of any single one is returned */
};

struct lestrade_match {
  u16 pos;				/* Position in the text measured in words */
  byte length;				/* Length in words */
  byte word_index;			/* Which word was matched or MATCH_PHRASE or MATCH_NEAR */
  s16 weight;				/* Weight of the match */
} PACKED;

enum {
  MATCH_PHRASE = 0x80,
  MATCH_NEAR
};

struct lestrade_query *lestrade_new_query(void);
void lestrade_free_query(struct lestrade_query *Q);
void lestrade_dump_query(struct lestrade_query *Q);
struct lestrade_expr *lestrade_new_op(struct lestrade_query *Q, uint op, struct lestrade_expr *l, struct lestrade_expr *r);
struct lestrade_expr *lestrade_new_match(struct lestrade_query *Q);
struct lestrade_atom *lestrade_add_atom(struct lestrade_query *Q, struct lestrade_expr *e, byte *word);

int lestrade_init_query(struct lestrade_query *Q);
int lestrade_eval_index(struct lestrade_query *Q, struct lestrade_index *idx);
int lestrade_eval_cards_init(struct lestrade_query *Q);
int lestrade_eval_card(struct lestrade_query *Q, struct card_attrs *attrs); 			/* Returns whether the card matched */
int lestrade_eval_cards(struct lestrade_query *Q, card_enumerator_t source, void *enum_priv);	/* Returns the number of matching cards */
int lestrade_eval_cards_cleanup(struct lestrade_query *Q);
void lestrade_get_results(struct lestrade_query *Q);

/* Find detailed positions of matches; the structure will be overwritten by subsequent calls */
int lestrade_eval_detailed_init(struct lestrade_query *Q);
struct lestrade_matches *lestrade_eval_detailed(struct lestrade_query *Q, struct card_attrs *attrs);
int lestrade_eval_detailed_cleanup(struct lestrade_query *Q);

/* Tracing options */

enum lestrade_trace {
  TRACE_GENERAL = 1,			/* Trace overall progress of query processing */
  TRACE_QUERY = 2,			/* Trace submitted queries */
  TRACE_PREPARE = 4,			/* Trace preparation phase */
  TRACE_WORDS = 8,			/* Trace word expansions */
  TRACE_FULLTEXT = 16,			/* Trace fulltext matching */
  TRACE_INDEX = 32,			/* Trace index matching */
};

#endif
